
import os
import urllib
import shutil
from gettext import gettext as _

import gobject
import gtk
import gnome
import gnomevfs
import wnck

from gimmie_globals import gimmie_is_panel_applet
from gimmie_base import Item, ItemSource, DisabledItemSource, Topic, gimmie_get_topic_for_uri
from gimmie_file import FileItem
from gimmie_recent import RecentlyUsed, RecentlyUsedOfMimeType
from gimmie_util import *
from gimmie_running import RunningItemSource, RunningItem
from gimmie_trash import trash_monitor
from gimmie_computer import DriveItem
from gimmie_tomboy import TomboySource


class RecentAggregate(ItemSource):
    '''
    This ItemSource subclass aggregates all the items from a list of
    ItemSources, by including the first Item encountered of a URI and
    filtering duplicates.
    '''
    def __init__(self, sources):
        ItemSource.__init__(self, name=_("Recently Used"), icon="stock_calendar")

        # Sources provide the real items we will display
        self.sources = sources
        for source in sources:
            self._listen_to_source(source)

    def _listen_to_source(self, source):
        source.connect("reload", lambda x: self.emit("reload"))
        source.connect("reload-item", lambda x, i: self.emit("reload"))

    def get_items(self):
        items = []

        # Find items matching recent uris
        for source in self.sources:
            # Block non-recent items the source might include like "Create New Note"
            items += [x for x in source.get_items() if x.get_timestamp() != 0]

        # Sort alphabetically
        #try:
        items.sort(key=lambda o: (o.get_name() and o.get_name().lower()))
        #except UnicodeDecodeError:
        #    items.sort(key=lambda o: (o.get_name() and o.get_name().decode().lower()))

        return items

    def do_reload(self):
        pass # Do nothing.


class RecentDocuments(RecentlyUsed):
    '''
    Recently used documents (non-launchers)
    '''
    def __init__(self, name = _("Recently Seen"), icon = "stock_calendar"):
        RecentlyUsed.__init__(self, name, icon)

    def include_recent_item(self, item):
        return not item.in_group("Launchers") \
               and item.get_mime_type() not in ("x-directory/normal",
                                                "x-tomboy/note", # FIXME: Is this needed?
                                                "application/octet-stream") \
               and urllib.splittype(item.get_uri())[0] not in ("ipp", "trash")


class RunningDocuments(RunningItemSource):
    '''
    A quick hack to look for DESKTOP_STARTUP_IDs containing URIs
    on Gimmie-launched running windows.  This only works for
    documents opened via the Documents window.
    '''
    def __init__(self):
        RunningItemSource.__init__(self, "Opened Documents")

    def get_bookmark_items(self):
        items = []
        # Make FileItem for all document bookmarks
        for uri, mime in bookmarks.get_bookmarks():
            if mime not in ("application/x-desktop", "gaim/buddy"):
                try:
                    items.append(FileItem(uri))
                except ValueError:
                    pass
        return items

    def get_item_for_uri(self, launch_uri):
        item = FileItem(launch_uri)
        if item.get_mimetype() in ("application/x-desktop",
                                   "gaim/buddy",
                                   "gaim/log",
                                   None):
            # Avoid showing launchers or conversations as running documents
            return None
        return item

    def do_reload(self):
        apps = []
        items = []
        bookmark_items = []

        # Make DesktopFileItems for all launcher bookmarks
        bookmark_items = self.get_bookmark_items()
        items.extend(bookmark_items)

        active_ws = self.screen.get_active_workspace()
        for win in self.screen.get_windows():
            if not win.is_on_workspace(active_ws):
                continue
            
            app = win.get_application()
            if not app or app in apps:
                continue
            apps.append(app)
            
            id = app.get_startup_id()
            if id and id.startswith("GIMMIE:"):
                try:
                    uri = id[len("GIMMIE:"):id.rfind("_TIME")]
                    timestamp = id[id.rfind("_TIME") + len("_TIME"):]

                    for b in bookmark_items:
                        if uri and (uri == b.get_uri() or uri == b.get_local_path()):
                            items[items.index(b)] = RunningItem(app, b)
                            break
                    else:
                        if uri:
                            item = self.get_item_for_uri(uri)
                            if item:
                                items.append(RunningItem(app, item))
                        #else:
                        #    items.append(RunningItem(app, None))
                except IndexError:
                    continue

        self.set_items(items)
        return False

    def handle_drag_data_received(self, selection, target_type):
        for uri in selection.get_uris():
            item = self.get_item_for_uri(uri)
            if item:
                bookmarks.add_bookmark(uri, item.get_mimetype())


class DirectorySource(ItemSource):
    def __init__(self, path, icon = "stock_folder"):
        self.path = os.path.expanduser(path)
        dirname = os.path.basename(self.path)
        ItemSource.__init__(self, dirname, icon=icon)

        self.monitor = FileMonitor(self.path)
        self.monitor.connect("changed", lambda x, file: self.emit("reload"))

    def get_enabled(self):
        return os.path.exists(self.path)

    def add_name(self, paths, dirname, names):
        for name in names:
            paths.append(os.path.join(dirname, name))

    def do_reload(self):
        paths = []
        os.path.walk(self.path, self.add_name, paths)
        self.set_items([FileItem(x) for x in paths])


class DownloadsDirectorySource(DirectorySource):
    def __init__(self):
        if os.environ.has_key("DOWNLOAD_PATH"):
            path = os.environ["DOWNLOAD_PATH"]
        else:
            path = "~/Downloads"
        DirectorySource.__init__(self, path, icon="stock_internet")


class AddSmartFolder(gtk.Dialog):
    '''
    Dialog to choose a stock Smart Folder to add to the sidebar,
    possibly prompting for some info (e.g. Tag names,
    modification date, etc).
    '''
    def __init__(self, parent):
        gtk.Dialog.__init__(self,
                            _("New Smart Folder"),
                            parent,
                            gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_NO_SEPARATOR,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))

    def get_sidebar_source(self):
        return None


class NewFromTemplate(gtk.FileChooserDialog):
    '''
    Dialog to create a new document from a template
    '''
    def __init__(self, parent, name):
        gtk.FileChooserDialog.__init__(self,
                                       _("New Document"),
                                       parent,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                        gtk.STOCK_SAVE, gtk.RESPONSE_ACCEPT))
        self.set_current_name(name)
        self.set_current_folder(os.path.expanduser("~/Desktop"))
        self.set_do_overwrite_confirmation(True)
        self.set_default_response(gtk.RESPONSE_ACCEPT)


class DocumentsTopic(Topic):
    '''
    Lists recently opened/edited documents from ~/.recently-used, and some other
    categories of files such as movies and music.  Placeholders for recent email
    attachments, marked emails, and browser downloads.  In the future categories
    displayed should be pluggable.
    '''
    def __init__(self):
        Topic.__init__(self, _("Documents"), uri="topic://Documents")
        self.set_running_source_factory(lambda: RunningDocuments())

        # Tomboy loads notes in a worker thread, so begin it early.
        self.tomboy_source = TomboySource()
        
        # Monitor the templates directory
        self.template_monitor = FileMonitor(os.path.expanduser("~/Templates"))
        self.template_monitor.open()

    def do_reload(self):
        downloads = DownloadsDirectorySource()
        
        ### FIXME: Prune to document types at the current zoom level?        
        source_list = [RecentAggregate([self.tomboy_source, downloads, RecentDocuments()]),
                       None,
                       #DisabledItemSource(_("Bookmarks"), "stock_bookmark"),
                       downloads,
                       DisabledItemSource(_("Attachments"), "stock_attach"),
                       None,
                       DisabledItemSource(_("Important Email"), "stock_mail"),
                       self.tomboy_source, # Tomboy
                       #RecentlyUsedOfMimeType(
                       #    _("Office Documents"),
                       #    "stock_new-presentation",
                       #    ["application/ms-powerpoint",
                       #     "application/vnd.oasis.opendocument.presentation",
                       #    "application/vnd.sun.xml.impress",
                       #    "application/x-abiword",
                       #     "text/plain",
                       #     "application/msword",
                       #     "application/vnd.ms-excel",
                       #     "application/x-gnumeric"]),
                       #RecentlyUsedOfMimeType(_("PDFs"),
                       #                          "gnome-pdf",
                       #                          ["application/pdf",
                       #                           "application/ps"
                       #                           "application/x-dvi"]),
                       RecentlyUsedOfMimeType(_("Music & Movies"),
                                              "gnome-mime-video",
                                              ["video/x-ms-wmv",
                                               "video/mpeg",
                                               "video/x-avi",
                                               "video/x-ms-asf",
                                               "video/x-mpeg",
                                               "video/x-ogg",
                                               "video/x-msvideo",
                                               "video/msvideo",
                                               "video/quicktime",
                                               "application/ogg"]),
                       DisabledItemSource(_("Photos"), "gthumb"),
                       ]

        self.set_sidebar_source_list(source_list)

    def get_hint_color(self):
        return gtk.gdk.color_parse("lightgreen")

    def _add_place_item(self,
                        menu,
                        name,
                        uri,
                        icon_name = "gnome-fs-directory"):
        menu_item = gtk.ImageMenuItem(name)
        menu_item.set_image(icon_factory.load_image(icon_name, gtk.ICON_SIZE_LARGE_TOOLBAR))
        menu_item.show()
        menu_item.connect("activate", lambda mi, uri: launcher.launch_uri(uri), uri)
        return menu_item

    def _add_places(self, menu):
        # Include all the mounted drives
        computer = gimmie_get_topic_for_uri("topic://Computer")
        device_source = computer.get_source_for_uri("source://Devices")
        for item in device_source.get_items():
            if isinstance(item, DriveItem) and item.get_is_mounted():
                menu.append(self._add_place_item(menu,
                                                 item.get_name(),
                                                 item.get_uri(),
                                                 item.get_icon(gtk.ICON_SIZE_LARGE_TOOLBAR)))

        sep = gtk.SeparatorMenuItem()
        sep.show()
        menu.append(sep)

        # Include all the Gtk bookmarks
        for uri, name, mime, icon in places.get_places():
            menu.append(self._add_place_item(menu, name, uri, icon))

        sep = gtk.SeparatorMenuItem()
        sep.show()
        menu.append(sep)

        # Add the trash
        if trash_monitor.is_empty():
            menu.append(self._add_place_item(menu, _("Trash"), "trash://", "gnome-fs-trash-empty"))
        else:
            menu.append(self._add_place_item(menu, _("Trash"), "trash://", "gnome-fs-trash-full"))

    def _template_changed(self, monitor, uri, btn):
        menu = gtk.Menu()
        self._add_templates(menu)
        try:
            btn.get_menu().destroy()
        except:
            pass
        btn.set_menu(menu)

    def _add_template_item(self,
                           menu,
                           name,
                           uri,
                           icon_name):
        menu_item = gtk.ImageMenuItem(name)
        menu_item.set_image(icon_factory.load_image(icon_name, gtk.ICON_SIZE_LARGE_TOOLBAR))
        menu_item.show()
        menu_item.connect("activate", self._show_new_from_template_dialog, name, uri)
        return menu_item

    def _add_templates(self, menu):
        # Get all the templates
        template_dir = os.path.expanduser("~/Templates")
	try:
        	templates = os.listdir(template_dir)
	except:
		templates = []

        for template in templates:
            item = FileItem(os.path.join(template_dir, template))
            menu.append(self._add_template_item(menu,
                                                 item.get_name(),
                                                 item.get_uri(),
                                                 item.get_icon(gtk.ICON_SIZE_LARGE_TOOLBAR)))
        if len(templates) == 0:
            empty_item = gtk.MenuItem(_("No templates available"))
            empty_item.set_sensitive(False)
            empty_item.show()
            menu.append(empty_item)

        sep = gtk.SeparatorMenuItem()
        sep.show()
        menu.append(sep)

        menu.append(self._add_template_item(menu,
                                            _("Empty File"),
                                            None,
                                            "stock_new"))

        sep = gtk.SeparatorMenuItem()
        sep.show()
        menu.append(sep)

        # Add a link to the templates directory
        menu.append(self._add_place_item(menu, _("Templates"), template_dir))

    def _show_new_from_template_dialog(self, b, name, uri):
        # Extract the template's file extension
        try:
            file_extension = name[name.rindex('.'):]
            file_name = name[:name.rindex('.')]
        except ValueError:
            file_extension = None
            file_name = name
        
        # Show The Dialog
        dlg = NewFromTemplate(b.get_toplevel(), file_name)
        if dlg.run() == gtk.RESPONSE_ACCEPT:
            file_uri = dlg.get_filename()

            # Add the file extension back unless the user did themselves
            if (file_extension != None and file_uri[file_uri.rfind('.'):] != file_extension):
                file_uri = "%s%s" % (file_uri, file_extension)

            # Create a new document from the template and display it
            try:
                if not uri:
                    # Create an empty file
                    f = open(file_uri, 'w')
                    f.close()
                else:
                    shutil.copyfile(uri, file_uri)
                launcher.launch_uri(file_uri)
            except IOError:
                pass

        dlg.destroy()

    def _show_add_smart_folder_dialog(self, b):
        dlg = AddSmartFolder(b.get_toplevel())
        if dlg.run() == gtk.RESPONSE_ACCEPT:
            source = dlg.get_sidebar_source()
            ### FIXME: Add to sources sidebar
            print "Adding!"
        dlg.destroy()

    def get_toolbar_items(self, tooltips):
        tools = []

        img = icon_factory.load_image("stock_new-template", gtk.ICON_SIZE_LARGE_TOOLBAR)
        btn = ToolMenuButton(img, _("New Document"))
        btn.set_tooltip(tooltips, _("Create a new document"))
        btn.set_is_important(True)
        
        self.template_monitor.connect("created", self._template_changed, btn)
        self.template_monitor.connect("deleted", self._template_changed, btn)
        self._template_changed(None, None, btn)
        tools.append(btn)

        '''
        img = icon_factory.load_image("stock_new-dir", gtk.ICON_SIZE_LARGE_TOOLBAR)
        btn = gtk.ToolButton(img, _("New Smart Folder"))
        btn.set_tooltip(tooltips, _("Add a new smart folder to the sidebar"))
        btn.set_is_important(True)
        btn.connect("clicked", self._show_add_smart_folder_dialog)
        tools.append(btn)
        '''

        img = icon_factory.load_image("gnome-fs-bookmark", gtk.ICON_SIZE_LARGE_TOOLBAR)
        btn = ToolMenuButton(img, _("Go"))
        btn.set_tooltip(tooltips, _("Open commonly used document locations"))
        btn.set_is_important(True)
        menu = gtk.Menu()
        self._add_places(menu)
        btn.set_menu(menu)
        tools.append(btn)

        ### FIXME: Include this?
        '''
        img = icon_factory.load_image("gnome-fs-trash-full", gtk.ICON_SIZE_LARGE_TOOLBAR)
        btn = gtk.ToolButton(img, _("Trash"))
        btn.set_tooltip(tooltips, _("Open the deleted files trashcan"))
        btn.set_is_important(True)
        btn.connect("clicked", lambda x: launcher.launch_uri("trash://"))
        tools.append(btn)
        '''

        if not gimmie_is_panel_applet():
            tools.append(None)

            img = icon_factory.load_image(gtk.STOCK_PRINT, gtk.ICON_SIZE_LARGE_TOOLBAR)
            btn = gtk.ToolButton(img, _("Printers"))
            btn.set_tooltip(tooltips, _("View printers available for document printing"))
            btn.set_is_important(True)
            btn.connect("clicked", lambda x: self._open_printers())
            tools.append(btn)

        return tools

    def _open_printers(self):
        topic = gimmie_get_topic_for_uri("topic://Computer")
        topicwin = topic.get_topic_window()
        topicwin.set_source_by_uri("source://Printers")
        topicwin.present()

    def get_context_menu_items(self):
        items = []

        img = icon_factory.load_image("stock_new-template", gtk.ICON_SIZE_LARGE_TOOLBAR)
        btn = gtk.ImageMenuItem(_("New Document"))
        btn.set_image(img)
        menu = gtk.Menu()
        self._add_templates(menu)
        btn.set_submenu(menu)
        btn.show_all()
        items.append(btn)

        img = icon_factory.load_image("gnome-fs-bookmark", gtk.ICON_SIZE_LARGE_TOOLBAR)
        btn = gtk.ImageMenuItem(_("Go"))
        btn.set_image(img)
        menu = gtk.Menu()
        self._add_places(menu)
        btn.set_submenu(menu)
        btn.show_all()
        items.append(btn)

        items.append(None)

        img = icon_factory.load_image(gtk.STOCK_PRINT, gtk.ICON_SIZE_LARGE_TOOLBAR)
        btn = gtk.ImageMenuItem(_("Printers"))
        btn.set_image(img)
        btn.connect("activate", lambda x: self._open_printers())
        btn.show_all()
        items.append(btn)
        
        return items
