
import datetime
import os
from gettext import gettext as _

import gobject
import gtk
import dbus

from gimmie_base import Item, ItemSource, DisabledItemSource, Topic
from gimmie_running import RunningItemSource
from gimmie_util import bookmarks, icon_factory, ToolMenuButton, ImplementMe
from gimmie_gaim import gaim_reader, gaim_dbus

try:
    from gimmie_gmail import gmail_reader
except ImportError, err:
    print "Unable to load GMail contact support:", err
    gmail_reader = None


#
#  People
#

class RecentConversations(ItemSource):
    '''
    A recent conversation source that supports filtering by last chat date.
    This uses the Gaim account reader so only chats with buddies in the buddy
    list will be displayed.

    This is a bug, but using the GaimReader makes things much faster, since we
    don\'t traverse the filesystem, and provides a single search result for the
    same buddy.
    '''
    def __init__(self, name, icon = None):
        ItemSource.__init__(self, name=name, icon=icon, uri="source://People/Recent")

    def do_reload(self):
        buddies = gaim_reader.get_all_buddies()
        buddies.sort(key=lambda o: o.get_name().lower())
        self.set_items(buddies)


class GmailSource(ItemSource):
    '''
    A list of Gmail contacts
    '''
    GMAIL_CONTACTS_UPDATE_INTERVAL = 3600000 # 1 hour in milliseconds
    
    def __init__(self, name, icon = None):
        ItemSource.__init__(self, name, icon, uri="source://People/Gmail", filter_by_date=False)
        self._DIRTY = True

        gmail_reader.connect("reload", lambda x: self.emit("reload"))

        # Set up timer to update contacts at regular intervals.
        gobject.timeout_add(GmailSource.GMAIL_CONTACTS_UPDATE_INTERVAL, lambda: self.do_reload())

    def get_items(self):
        if self._DIRTY:
            self._reload()
        return ItemSource.get_items(self)

    def _reload(self):
        items = [gmail_reader.get_account_settings()]
        contacts = gmail_reader.get_contacts()
        if contacts:
            items += contacts
        
        self.set_items(items)

        self._DIRTY = False

    def do_reload(self):
        # We call _reload in get_items to load on-demand.
        self._DIRTY = True

        # Return True to continue timeout
        return True


class GroupSource(ItemSource):
    '''
    A list of all people in a GaimGroup, filtered by active accounts
    '''
    def __init__(self, group, icon = "gaim"):
        uri = "source://People/Group/%s" % group.get_name()
        ItemSource.__init__(self, group.get_name(), icon=icon, uri=uri, filter_by_date=False)
        self.group = group

    def do_reload(self):
        buddies = [b for b in self.group.get_buddies() if b.get_account().get_auto_login()]
        buddies.sort(key=lambda o: o.get_name().lower())
        self.set_items(buddies)


class EverybodySource(ItemSource):
    '''
    A list of "everyone", filtered by active accounts
    '''
    def __init__(self, name, icon = None):
        ItemSource.__init__(self, name, icon, uri="source://People/Everybody", filter_by_date=False)

        # Listen for changes in the Gaim files
        gaim_reader.connect("reload", lambda x: self.emit("reload"))

    def do_reload(self):
        everyone = gaim_reader.get_all_buddies()
        everyone.sort(key=lambda o: o.get_name().lower())
        self.set_items(everyone)


class OnlineBuddySource(ItemSource):
    '''
    A list of all currently online buddies.
    '''
    def __init__(self, name, icon = None):
        ItemSource.__init__(self, name, icon, uri="source://People/Online", filter_by_date=False)
        self._DIRTY = True

        # Listen for changes in the Gaim files.  This isn't a good indicator of
        # presence changes, but it works sometimes.
        gaim_reader.connect("reload", lambda x: self.emit("reload"))

    def get_items(self):
        if self._DIRTY:
            self._reload()
        return ItemSource.get_items(self)

    def _reload(self):
        online = []
        try:
            online = [x for x in gaim_reader.get_all_buddies() if x.get_is_online()]
            online.sort(key=lambda o: o.get_name().lower())
        except (dbus.DBusException, TypeError):
            print " !!! Error accessing Gaim2 D-BUS interface.  Is Gaim2 running?"
        self.set_items(online)
        self._DIRTY = False

    def do_reload(self):
        # We call _reload in get_items to load on-demand.
        self._DIRTY = True


class RunningChat(Item):
    '''
    A simple Item subclass that focuses the window passed to init on open.
    '''
    def __init__(self, buddy, win):
        Item.__init__(self,
                      uri=buddy.get_uri(),
                      timestamp=buddy.get_timestamp(),
                      mimetype="gaim/conversation",
                      icon=None)
        self.buddy = buddy
        self.win = win

        # Listen for hint changes.
        self.win.connect("state-changed", self._win_state_changed)

    def _win_state_changed(self, win, changemask, newstate):
        ### FIXME: We seem to miss events if we filter, so always emit
        # 1024 == WNCK_WINDOW_STATE_DEMANDS_ATTENTION
        #if changemask & 1024:
        self.emit("reload")
            
    def get_icon(self, icon_size):
        return self.buddy.get_icon(icon_size)

    def get_name(self):
        return self.buddy.get_name()

    def get_is_opened(self):
        return True

    def get_demands_attention(self):
        return self.win.needs_attention()

    def do_open(self):
        self.win.activate(gtk.get_current_event_time())

    def handle_drag_data_received(self, selection, target_type):
        for uri in selection.get_uris():
            if uri.lower().startswith("aim:goim?screenname="):
                bookmarks.add_bookmark(uri, "gaim/buddy")


class RunningChats(RunningItemSource):
    '''
    A quick hack to use Gaim window titles to determine currently active chats.
    This only works if you turn off tabbed chat mode and enable the remote
    control plugin.
    '''
    def __init__(self):
        RunningItemSource.__init__(self, "Active Conversations")

        # Listen for changes in the Gaim files
        gaim_reader.connect("reload", lambda x: self.emit("reload"))

        ### Gaim2 D-BUS interface has no way to signal changes, so just poll
        #gobject.timeout_add(1000, lambda: self.emit("reload") or True)

    def _get_screenname_from_uri(self, uri):
        if uri.lower().startswith("aim:goim?screenname="):
            return uri[len("aim:goim?screenname="):].strip("\"'") # Trim quotemarks
        return None

    def _get_buddy_for_screenname(self, screenname):
        screenname = screenname.lower()
        buddies = [x for x \
                   in gaim_reader.get_all_buddies() \
                   if screenname in (x.get_screenname().lower(), x.get_alias().lower())]
        if not buddies:
            return None

        # FIXME: Filter for online accounts.  Just use auto-login for now.
        if len(buddies) > 1:
            auto_buddies = [x for x in buddies if x.get_account().get_auto_login()]
            if auto_buddies:
                buddies = auto_buddies

        # FIXME: This arbitrarily picks the first one
        return buddies[0]

    def get_bookmark_items(self):
        items = []
        for uri, mime in bookmarks.get_bookmarks(["gaim/buddy"]):
            buddy = self.get_item_for_uri(uri)
            if buddy:
                items.append(buddy)
        return items

    def get_item_for_uri(self, launch_uri):
        try:
            screenname = self._get_screenname_from_uri(launch_uri)
            if screenname == None:
                return None
            buddy = self._get_buddy_for_screenname(screenname)
            return buddy
        except (ValueError, IndexError):
            return None

    def do_reload(self):
        items = []

        # Get GaimBuddy items for all people bookmarks
        bookmark_items = self.get_bookmark_items()
        items.extend(bookmark_items)

        ### FIXME: Use Gaim2 D-BUS to list open conversations.  More work needed here.
        #try:
        #    convs = gaim_dbus.GaimGetConversations()
        #    conv_names = [gaim_dbus.GaimConversationGetName(x) for x in convs]
        #    conv_buddies = [self._get_buddy_for_screenname(x) for x in conv_names]
        #
        #    items += [b for b in conv_buddies if b not in bookmark_items]
        #    if items == self.items:
        #        print "HALTING EMISSION!"
        #        self.emit_stop_by_name("reload")
        #    else:
        #        self.set_items(items)
        #    return True
        #except (dbus.DBusException, TypeError):
        #    pass

        active_ws = self.screen.get_active_workspace()
        for win in self.screen.get_windows():
            if not win.is_on_workspace(active_ws):
                continue

            if win.get_class_group().get_name() != "gaim":
                continue

            screenname = win.get_name().lower()
            if screenname[0] == '(' and screenname[-1] == ')': # Strip offline indicator
                screenname = screenname[1:-1]

            for b in bookmark_items:
                if screenname in (b.get_screenname().lower(), b.get_alias().lower()):
                    items[items.index(b)] = RunningChat(b, win)
                    break
            else:
                buddy = self._get_buddy_for_screenname(screenname)
                if buddy:
                    items.append(RunningChat(buddy, win))

            # FIXME: Figure out proper tabbed conversation support.
            # Look for a new buddy match when the window title changes.
            if not hasattr(win, "gimmie_name_changed_slot"):
                win.gimmie_name_changed_slot = win.connect("name-changed",
                                                           lambda x: self.emit("reload"))

        self.set_items(items)
        return False

    def handle_drag_data_received(self, selection, target_type):
        for uri in selection.get_uris():
            item = self.get_item_for_uri(uri)
            if item:
                bookmarks.add_bookmark(uri, item.get_mimetype())


class PeopleTopic(Topic):
    '''
    Lists recently talked to people, people online now, and all contacts, as
    well as listing out Gaim groups.  Placeholders for future
    Gmail/Friendster/LinkedIn contact integration.
    '''
    def __init__(self):
        Topic.__init__(self, _("People"), uri="topic://People")

        self.set_running_source_factory(lambda: RunningChats())

        # Reload the sidebar when Gaim's files change
        gaim_reader.connect("reload", lambda x: self.emit("reload"))

    def do_reload(self):
        # FIXME: Prune for groups at the current zoom level?
        source_list = [RecentConversations(_("Recent People"), "im"),
                       OnlineBuddySource(_("Online Now"), "gaim"),
                       EverybodySource(_("Everybody"), "gnome-globe"),
                       None]

        if gmail_reader:
            source_list += [GmailSource(_("Gmail"), "stock_mail")]
        else:
            source_list += [DisabledItemSource(_("GMail"), "stock_mail")]
            
        source_list += [DisabledItemSource(_("Friendster"), "stock_people"),
                        None]

        for group in gaim_reader.get_groups():
            for buddy in group.get_buddies():
                # Only include groups which have at least one active buddy
                if buddy.get_account().get_auto_login():
                    source = GroupSource(group)
                    source_list.append(source)
                    break

        self.set_sidebar_source_list(source_list)

    def get_hint_color(self):
        return gtk.gdk.color_parse("pink")

    def _show_accounts_editor(self):
        ImplementMe()

    def make_availability_menu(self):
        menu = gtk.Menu()

        item = gtk.RadioMenuItem(None, _("Available"))
        item.set_active(True)
        item.show()
        menu.append(item)

        item = gtk.RadioMenuItem(item, _("Offline"))
        item.show()
        menu.append(item)

        away_menu = gtk.Menu()
        away_item = gtk.MenuItem(_("Away Messages..."))
        away_item.show()
        away_menu.append(away_item)

        item = gtk.MenuItem(_("Away"))
        item.show()
        item.set_submenu(away_menu)
        menu.append(item)

        item = gtk.SeparatorMenuItem()
        item.show()
        menu.append(item)

        item = gtk.MenuItem(_("Accounts..."))
        item.connect("activate", lambda w: self._show_accounts_editor())
        item.show()
        menu.append(item)

        item = gtk.SeparatorMenuItem()
        item.show()
        menu.append(item)

        for acct in gaim_reader.get_accounts():
            item = gtk.CheckMenuItem(acct.get_name())
            if acct.get_auto_login():
                item.set_active(True)
            item.show()
            menu.append(item)

        # FIXME: Everything here isn't implemented yet.
        for mi in menu:
            mi.set_sensitive(False)
        return menu

    def _gaim_get_default_account_obj(self):
        acct = gaim_dbus.GaimAccountsFindConnected("", "")
        if not acct:
            acct = gaim_dbus.GaimAccountsFindAny("", "")
        return acct

    def _add_person(self, w):
        if gaim_dbus:
            try:
                acct = self._gaim_get_default_account_obj()
                gaim_dbus.GaimBlistRequestAddBuddy(acct, "", "", "")
            except dbus.DBusException:
                print " !!! Gaim2 D-BUS interface not available.  Is Gaim2 running?"

    def _add_group(self, w):
        if gaim_dbus:
            try:
                gaim_dbus.GaimBlistRequestAddGroup()
            except dbus.DBusException:
                print " !!! Gaim2 D-BUS interface not available.  Is Gaim2 running?"

    def get_toolbar_items(self, tooltips):
        tools = []

        img = icon_factory.load_image("stock_person", gtk.ICON_SIZE_LARGE_TOOLBAR)
        btn = gtk.ToolButton(img, _("New Person"))
        btn.set_tooltip(tooltips, _("Add a new contact person"))
        btn.set_is_important(True)
        btn.connect("clicked", self._add_person)
        tools.append(btn)

        ### Groups are lame
        '''
        img = icon_factory.load_image("stock_people", gtk.ICON_SIZE_LARGE_TOOLBAR)
        btn = gtk.ToolButton(img, _("New Group"))
        btn.set_tooltip(tooltips, _("Add a new group of people to the sidebar"))
        btn.set_is_important(True)
        btn.connect("clicked", self._add_group)
        tools.append(btn)
        '''

        img = icon_factory.load_image(gtk.STOCK_YES, gtk.ICON_SIZE_BUTTON)
        btn = ToolMenuButton(img, _("Available"))
        btn.set_tooltip(tooltips, _("Set your online status"))
        btn.set_is_important(True)
        btn.set_menu(self.make_availability_menu())
        tools.append(btn)

        return tools

    def get_context_menu_items(self):
        items = []

        img = icon_factory.load_image(gtk.STOCK_YES, gtk.ICON_SIZE_LARGE_TOOLBAR)
        btn = gtk.ImageMenuItem(_("Available"))
        btn.set_image(img)
        btn.set_submenu(self.make_availability_menu())
        btn.show_all()
        items.append(btn)
        
        return items

