
from gettext import gettext as _

import gobject
import gtk
import wnck

from gimmie_base import Item, ItemSource
from gimmie_util import bookmarks, icon_factory


class RunningItem(Item):
    '''
    A wrapper around an Item and a WnckClassGroup or WnckApplication.
    '''
    def __init__(self, app_group, item = None):
        assert app_group, "WnckClassGroup app_group cannot be None"
        
        if item:
            Item.__init__(self,
                          uri=item.get_uri(),
                          timestamp=item.get_timestamp(),
                          mimetype=item.get_mimetype())
        else:
            Item.__init__(self,
                          uri="win-group://%s" % app_group.get_name(),
                          mimetype="x-unknown")

        self.item = item
        self.app_group = app_group

        # Listen for hint changes.
        for win in app_group.get_windows():
            win.connect("state-changed", self._win_state_changed)

    def get_name(self):
        if self.item:
            return self.item.get_name()
        else:
            return self.app_group.get_name()

    def get_comment(self):
        if self.item:
            return self.item.get_comment()
        else:
            return ""

    def get_icon(self, icon_size):
        if self.item:
            return self.item.get_icon(icon_size)
        else:
            ### FIXME: Support appgroups and applications?
            
            ### FIXME: Gaim has a tendency to set the group leader icon to the latest buddy icon
            if self.app_group.get_res_class() == "Gaim":
                icon = "/usr/share/pixmaps/gaim.png"
            else:
                icon = self.app_group.get_icon()
            return icon_factory.load_icon(icon, icon_size)
            
    def do_open(self):
        screen = wnck.screen_get(0)
        opened_one = False

        app_wins = self.app_group.get_windows()
        active_ws = screen.get_active_workspace()

        # Sorted by stacking order
        visible_wins = [win for win
                        in screen.get_windows_stacked()
                        if win in app_wins and win.is_on_workspace(active_ws)]
        if not visible_wins and self.item:
            self.item.open()
            return

        if screen.get_active_window() in visible_wins:
            # Already active
            if len(visible_wins) > 1:
                # This group has the active win, so cycle the topmost active window down
                print " *** Raising window: ", visible_wins[0].get_name()
                visible_wins[0].activate(gtk.get_current_event_time())
            return
        else:
            # Not currently active, so raise them all
            for win in visible_wins:
                win.activate(gtk.get_current_event_time())
            return

        raise Exception("No application windows on this desktop, and no Item to launch")

    def get_is_opened(self):
        if self.app_group and self.app_group.get_windows():
            return True
        return (self.item and self.item.get_is_opened()) or False

    def get_is_active(self):
        if self.app_group:
            for win in self.app_group.get_windows():
                if win.is_active():
                    return True
        return (self.item and self.item.get_is_active()) or False
 
    def get_can_pin(self):
        if self.item:
            return self.item.get_can_pin()
        return False

    def set_screen_position(self, x, y, w, h):
        '''
        Set the _NET_WM_ICON_GEOMETRY window manager hint on all the
        app-group\'s windows so that they will minimize onto this item\'s area.
        '''
        for win in self.app_group.get_windows():
            win.set_icon_geometry(x, y, w, h)

    def get_tooltip(self):
        if self.item:
            return self.item.get_tooltip()
        else:
            return self.app_group.get_name()

    def _win_state_changed(self, win, changemask, newstate):
        ### FIXME: We seem to miss events if we filter, so always emit
        # 1024 == WNCK_WINDOW_STATE_DEMANDS_ATTENTION
        #if changemask & 1024:
        self.emit("reload")

    def get_is_user_visible(self):
        workspace = wnck.screen_get_default().get_active_workspace()
        for win in self.app_group.get_windows():
            if win.is_visible_on_workspace(workspace) and \
                   win.get_window_type() in (0, 7, 8):
                # (Normal, Utility, Splashscreen) window
                return True
        else:
            if self.item:
                return self.item.get_is_user_visible()
            return False

    def get_demands_attention(self):
        for win in self.app_group.get_windows():
            if win.needs_attention():
                return True
        return False

    def handle_drag_data_received(self, selection, target_type):
        if self.item:
            self.item.handle_drag_data_received(selection, target_type)


class RunningItemSource(ItemSource):
    def __init__(self, name):
        ItemSource.__init__(self, name, None)

        self.screen = wnck.screen_get(0)
        self.timeout_id = None

        self.screen.connect("window-opened", lambda x, y: self.queue_reload_maybe(y))
        self.screen.connect("window-closed", lambda x, y: self.queue_reload_maybe(y))
        self.screen.connect("application-opened", lambda x, y: self.queue_reload())
        self.screen.connect("application-closed", lambda x, y: self.queue_reload())
        self.screen.connect("active-workspace-changed", lambda x: self.queue_reload())

        bookmarks.connect("reload", lambda x: self.emit("reload"))
    
    def get_bookmark_items(self):
        raise NotImplementedError

    def get_item_for_uri(self, launch_uri):
        raise NotImplementedError

    def do_reload(self):
        raise NotImplementedError

    def queue_reload(self):
        if self.timeout_id:
            gobject.source_remove(self.timeout_id)
        self.timeout_id = gobject.timeout_add(50, lambda: self.emit("reload"))

    def queue_reload_maybe(self, win):
        # ignore dock windows (wnck bindings are broken)
        if win.get_window_type() != 2:
            self.queue_reload()
